# -*- coding: utf-8 -*-
'film_distance_relaxation': False,   # if True, sets relaxXYZ="FFT" for all atoms
'force_criterion': 0.049,            # Sets the threshold of the largest force
'relax_iter': 5                      # Maximum number of optimization iterations
