# -*- coding: utf-8 -*-
{'atomtypes': [[
  {'atomic_number': 4, 'coreconfig': '(1s1/2)', 'element': 'Be', 'natoms': 1,
   'species': 'Be_corehole1', 'stateOccupation': [
      {'(1s1/2)': ['1.00000000', '.50000000']},
      {'(2p1/2)': ['.50000000', '.00000000']}], 'valenceconfig': '(2s1/2) (2p1/2)'},
  {'atomic_number': 4, 'coreconfig': '[He]', 'element': 'Be', 'natoms': 1,
   'species': 'Be-2', 'stateOccupation': [{'(2p1/2)': ['.00000000', '.00000000']}],
   'valenceconfig': '(2s1/2) (2p1/2)'},
  {'atomic_number': 4, 'coreconfig': '[He]', 'element': 'Be', 'natoms': 1,
   'species': 'Be-2', 'stateOccupation': [{'(2p1/2)': ['.00000000', '.00000000']}],
   'valenceconfig': '(2s1/2) (2p1/2)'},
  {'atomic_number': 4, 'coreconfig': '[He]', 'element': 'Be', 'natoms': 1,
   'species': 'Be-2', 'stateOccupation': [{'(2p1/2)': ['.00000000', '.00000000']}],
   'valenceconfig': '(2s1/2) (2p1/2)'}]], 'bandgap': [0.0004425914],
  'bandgap_units':'eV', 'binding_energy': [53.57027767044], 'corehole_type': 'valence',
  'binding_energy_units': 'eV', 'binding_energy_convention': 'negativ',
  'coreholes_calculated': 'Be1s', 'coreholes_calculated_details': '', 'coresetup': [],
  'errors': [], 'fermi_energy': [0.3138075709], 'fermi_energy_unit': 'eV',
  'reference_bandgaps': [0.0225936434], 'reference_coresetup': [],
  'successful': true, 'total_energy_all': [-1554.08485250996],
  'total_energy_all_units': 'eV', 'total_energy_ref': [-1607.6551301804],
  'total_energy_ref_units': 'eV', 'warnings': [], 'hints': [],
  'weighted_binding_energy': [107.14055534088], 'weighted_binding_energy_units': 'eV',
  'workflow_name': 'fleur_corehole_wc', 'workflow_version': '0.3.2'
}
