# -*- coding: utf-8 -*-
'beta': {'all': 1.57079},                   # see the description below
'q_vectors': {'label': [0.0, 0.0, 0.0],     # sets q_points to calculate
              'label2': [0.125, 0.0, 0.0]
              }
'suppress_symmetries': False                # True if use no symmetries
