# -*- coding: utf-8 -*-
'element': '',                          # detemines for which element to calculate
                                        # the crystal field coefficients
'rare_earth_analogue': True,            # True if analogue calculation should be used
'analogue_element': 'Y',                # Which element to use for the analogue structure
'replace_all': True,                    # Whether to replace all atoms for the analogue in one structure
'soc_off': True,                        # if True the socscale is set to 0.0 for the 4f site
'convert_to_stevens': True              # if True the coefficients are converted to the stevens convention
                                        # A_lm<r^l>
